unit fdplain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Menus, ExtCtrls, StdCtrls;


type
  EInvalidPictureFile = class(Exception);


const
  MaxXConnectionPoints = 5; { a type has 4 grip points on X-as }

type
  TXConnectionPoint = 0..(MaxXConnectionPoints - 1);
  TYConnectionPoint = (Ytop, Ybottom);

const
  DefaultHeight = 40;
  DefaultWidth = 90;

type
  TXplainType = class(TPanel)
  public
    TypesAbove,
    TypesUnder: TStringList;
    OrigLeft,
    OrigTop: integer;
    constructor Create(AOwner: TComponent); override;
    function  CalculateXConnectionPoint(X: integer): TXConnectionPoint;
    function  CalculateYConnectionPoint(Y: integer): TYConnectionPoint;
    procedure ConnectTo(OtherType: TXplainType; XA,YA, XB,YB: integer);
    procedure RemoveConnection(OtherType: TXplainType);
  end;

type
  TLine = array[1..2] of TPoint;

type
  TXPlainConnection = class(TGraphicControl)
  protected
    procedure Paint; override;
  public
    Type1,
    Type2: TXplainType;
    X1ConnectionPoint,
    X2ConnectionPoint: integer;
    Y1ConnectionPoint,
    Y2ConnectionPoint: TYConnectionPoint;
(*
    Delta1,
    Delta2: TPoint;
    Points: TLine;
*)
    Selected: Boolean;
    function IsSelfReference: Boolean;
    function IsSubType: Boolean;
    procedure RemoveLinks;
    function X1: integer;
    function Y1: integer;
    function X2: integer;
    function Y2: integer;
  end;

type
  TXCanvasForm = class(TForm)
    MainMenu: TMainMenu;
    File1: TMenuItem;
    New1: TMenuItem;
    Open1: TMenuItem;
    Save1: TMenuItem;
    SaveAs1: TMenuItem;
    Object1: TMenuItem;
    Type1: TMenuItem;
    Base1: TMenuItem;
    N1: TMenuItem;
    Exit1: TMenuItem;
    XYPos: TLabel;
    TypePopupMenu: TPopupMenu;
    Name1: TMenuItem;
    Connect1: TMenuItem;
    Schema1: TMenuItem;
    Move1: TMenuItem;
    Movetoback1: TMenuItem;
    N2: TMenuItem;
    Delete1: TMenuItem;
    procedure Exit1Click(Sender: TObject);
    procedure Type1Click(Sender: TObject);
    procedure Panel1MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure Panel1MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure Panel1MouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure Name1Click(Sender: TObject);
    procedure Connect1Click(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure Save1Click(Sender: TObject);
    procedure SaveTypeAsTexFile;
    procedure OpenTexFile;
    procedure SaveAs1Click(Sender: TObject);
    procedure Open1Click(Sender: TObject);
    procedure New1Click(Sender: TObject);
    procedure Move1Click(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure Movetoback1Click(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure Delete1Click(Sender: TObject);
  private
    { Private declarations }
    LastPoints: TLine;
    HaveLastPoints: Boolean;
    ClickedType,
    DraggedType,
    SelectedType: TXplainType;
    ClickOrigin,
    DragOrigin,
    ConnectOrigin,
    MovePoint: TPoint;
    Mode: (mAddTypes, mDragType, mConnectTypes, mMoveSchema);
    Types: TStringList;
    Connections: TStringList;
    SelectedConnection: TXPlainConnection;
    FFileName: string;
    Grid: TPoint;
    Changed: Boolean;
    procedure RegisterConnection(
      Type1,Type2: TXplainType;
      X1ConnectionPoint: TXConnectionPoint;
      Y1ConnectionPoint: TYConnectionPoint;
      X2ConnectionPoint: TXConnectionPoint;
      Y2ConnectionPoint: TYConnectionPoint);
    procedure ClearLine;
    procedure MoveBehind(AType: TXplainType);
    procedure SetSelectedType(AType: TXplainType);
    procedure RemoveConnection(AConnection: TXPlainConnection);
    procedure RemoveType(AType: TXplainType);
    procedure SetFileName(AFileName: string);
  public
    { Public declarations }
    property FileName: string read FFileName write SetFileName;
  end;

var
  XCanvasForm: TXCanvasForm;

implementation

{$R *.DFM}


const
  TypeFormat = '\parbox{27mm}{\centering \sf ';



function UserWants(const Question: string): TModalResult;
begin
  Result := MessageDlg(Question, mtConfirmation, mbYesNoCancel, 0);
end;


{ TXplainType }

constructor TXplainType.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  BevelInner := bvLowered;
  BevelOuter := bvRaised;
  BorderStyle := bsNone;
  //BevelInner := bvNone;
  //BevelOuter := bvNone;
  //BorderStyle := bsSingle;
  Caption := 'type1';
  ParentColor := True;
  //Color := clWhite;
  Height := DefaultHeight;
  Width := DefaultWidth;
  TypesAbove := TStringList.Create;
  TypesUnder := TStringList.Create;
end;

function TXplainType.CalculateXConnectionPoint(X: integer): TXConnectionPoint;
begin
  Result := Round( (X / Width) * (MaxXConnectionPoints-1));
end;

function TXplainType.CalculateYConnectionPoint(Y: integer): TYConnectionPoint;
begin
  if Y > Height div 2
    then
      Result := Ybottom
    else
      Result := Ytop;
end;

procedure TXplainType.ConnectTo(OtherType: TXplainType; XA,YA, XB,YB: integer);
begin
  if Top <= OtherType.Top
    then  begin
      TypesUnder.AddObject(OtherType.Caption, OtherType);
      OtherType.TypesAbove.AddObject(Caption, Self);
    end
    else  begin
      TypesAbove.AddObject(OtherType.Caption, OtherType);
      OtherType.TypesUnder.AddObject(Caption, Self);
    end;
  XCanvasForm.RegisterConnection(
        Self, OtherType,
        CalculateXConnectionPoint(XA),
        CalculateYConnectionPoint(YA),
        OtherType.CalculateXConnectionPoint(XB),
        OtherType.CalculateYConnectionPoint(YB));
end;

procedure TXplainType.RemoveConnection(OtherType: TXplainType);
begin
  if Top = OtherType.Top then
  begin
  { circular link }
    TypesUnder.Delete(TypesUnder.IndexOf(OtherType.Caption));
    TypesAbove.Delete(TypesAbove.IndexOf(OtherType.Caption));
  end
  else
  begin
    if Top < OtherType.Top then
    begin
      TypesUnder.Delete(TypesUnder.IndexOf(OtherType.Caption));
    end
    else
    begin
      TypesAbove.Delete(TypesAbove.IndexOf(OtherType.Caption));
    end;
  end;
end;


{ TXplainConnection }

function TXplainConnection.IsSelfReference: Boolean;
begin
  Result := Type1 = Type2;
end;

function TXplainConnection.IsSubType: Boolean;
begin
  Result := X1ConnectionPoint in [0,4];
end;

procedure TXplainConnection.Paint;
var
  X1,Y1,X2,Y2,X3,Y3,X4,Y4: integer;
begin
  if Selected then
    Canvas.Pen.Color := clGreen;
  if IsSelfReference
    then  begin
      X1 := Self.X1;
      Y1 := Type1.Top + (Type1.Height div 2);
      if X1ConnectionPoint <= 2
        then
          X2 := Type1.Left - (X1 - Type1.Left)
        else
          X2 := (Type1.Left + Type1.Width) + ((Type1.Left + Type1.Width) - X1);
      Y2 := Type1.Top + (Type1.Height div 2);
      X3 := Self.X1;
      if Self.Y2 > Self.Y1
        then  Y3 := Self.Y2
        else  Y3 := Self.Y1;
      X4 := Self.X1;
      if Self.Y2 > Self.Y1
        then  Y4 := Self.Y1
        else  Y4 := Self.Y2;
      if X1ConnectionPoint <= 2
        then
          Canvas.Arc(X1,Y1-40, X2,Y2+40, X4,Y4, X3,Y3)
        else
          Canvas.Arc(X1,Y1-40, X2,Y2+40, X3,Y3, X4,Y4);
    end
    else  begin
      Canvas.MoveTo(Self.X1, Self.Y1);
      Canvas.LineTo(Self.X2, Self.Y2);
    end;
  if Selected then
    Canvas.Pen.Color := clBlack;
end;

procedure TXplainConnection.RemoveLinks;
begin
  Type1.RemoveConnection(Type2);
  if Type1 <> Type2 then
    Type2.RemoveConnection(Type1);
end;

function TXplainConnection.X1: integer;
begin
  Result := Type1.Left + Round((Type1.Width / (MaxXConnectionPoints-1)) * X1ConnectionPoint);
end;

function TXplainConnection.Y1: integer;
begin
  case Y1ConnectionPoint of
    Ytop: Result := Type1.Top;
    Ybottom: Result := (Type1.Top + Type1.Height);
  end;
end;

function TXplainConnection.X2: integer;
begin
  Result := Type2.Left + Round((Type2.Width / (MaxXConnectionPoints-1)) * X2ConnectionPoint);
end;

function TXplainConnection.Y2: integer;
begin
  case Y2ConnectionPoint of
    Ytop: Result := Type2.Top;
    Ybottom: Result := (Type2.Top + Type2.Height);
  end;
end;


{ TXCanvasForm }

procedure TXCanvasForm.Exit1Click(Sender: TObject);
begin
  Close;
end;

procedure TXCanvasForm.Type1Click(Sender: TObject);
var
  AType: TXplainType;
  TypeName: string;
begin
  TypeName := 'type1';
  if InputQuery('New type name', 'name', TypeName) then  begin
    AType := TXplainType.Create(Self);
    AType.Caption := TypeName;
    AType.Parent := Self;
    AType.OnMouseDown := Panel1MouseDown;
    AType.OnMouseMove := Panel1MouseMove;
    AType.OnMouseUp := Panel1MouseUp;
    AType.PopUpMenu := TypePopupMenu;
    Types.AddObject(TypeName, AType);
    SetSelectedType(AType);
    Changed := True;
  end;
end;

procedure TXCanvasForm.Panel1MouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  ClickedType := TXplainType(Sender);
  ClickOrigin.X := X;
  ClickOrigin.Y := Y;
  if Sender is TXplainType then  begin
    case Mode of
      mAddTypes: if Button = mbLeft then  begin
          Mode := mDragType;
          DraggedType := ClickedType;
          DragOrigin.X := X;
          DragOrigin.Y := Y;
        end;
      mConnectTypes: begin
          ClearLine;
          SelectedType.ConnectTo(TXplainType(Sender), ConnectOrigin.X, ConnectOrigin.Y, X,Y);
          Changed := True;
        end;
    end; { of case }
  end;
end;

procedure TXCanvasForm.Panel1MouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
begin
  case Mode of
    mDragType:
      begin
        XYPos.Caption := IntToStr(DraggedType.Left) + ',' + IntToStr(DraggedType.Top);
        DraggedType.Left := DraggedType.Left+(((X-DragOrigin.X) div Grid.X) * Grid.X);
        DraggedType.Top := DraggedType.Top+(((Y-DragOrigin.Y) div Grid.Y) * Grid.Y);
      end;
    mConnectTypes: begin
        FormMouseMove(Sender, Shift, TPanel(Sender).Left+X, TPanel(Sender).Top+Y);
      end;
  end;
end;

procedure TXCanvasForm.Panel1MouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  case Mode of
    mDragType:
      begin
        SetSelectedType(DraggedType);
        DraggedType := nil;
        Refresh;
      end;
    mAddTypes: SetSelectedType(ClickedType);
    mConnectTypes: HaveLastPoints := False;
  end;
  Mode := mAddTypes;
end;

procedure TXCanvasForm.FormKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  case Key of
    vk_Return: begin
        if (SelectedType <> nil) and (Shift = [ssAlt]) then  begin
          Name1Click(Sender);
        end;
      end;
    vk_Escape: begin
        Mode := mAddTypes;
        ClearLine;
      end;
    vk_Delete:
      begin
        if (SelectedConnection <> nil) and
           (UserWants('Delete this connection?') = mrYes) then
        begin
          RemoveConnection(SelectedConnection);
          Refresh;
        end;
        if (SelectedType <> nil) and
           (UserWants('Delete this type?') = MrYes) then
        begin
          RemoveType(SelectedType);
          Refresh;
        end;
      end;
  end;
end;

procedure TXCanvasForm.Name1Click(Sender: TObject);
var
  s: string;
begin
  s := SelectedType.Caption;
  if InputQuery('Type name', 'Name', s) then  begin
    SelectedType.Caption := s;
    Changed := True;
  end;
end;

procedure TXCanvasForm.Connect1Click(Sender: TObject);
begin
  Mode := mConnectTypes;
  ConnectOrigin := ClickOrigin;
end;

procedure TXCanvasForm.RegisterConnection(
  Type1,Type2: TXplainType;
  X1ConnectionPoint: TXConnectionPoint;
  Y1ConnectionPoint: TYConnectionPoint;
  X2ConnectionPoint: TXConnectionPoint;
  Y2ConnectionPoint: TYConnectionPoint);
var
  Connection: TXPlainConnection;
begin
  Connection := TXPlainConnection.Create(Self);
  Connection.Parent := Self;
  Connection.OnMouseDown := FormMouseDown;
  Connection.OnMouseMove := FormMouseMove;
  Connection.Type1 := Type1;
  Connection.Type2 := Type2;
  Connection.X1ConnectionPoint := X1ConnectionPoint;
  Connection.Y1ConnectionPoint := Y1ConnectionPoint;
  Connection.X2ConnectionPoint := X2ConnectionPoint;
  Connection.Y2ConnectionPoint := Y2ConnectionPoint;
  Connection.Align := alClient;
  Connections.AddObject('', Connection);
(*
  Canvas.Pen.Mode := pmNotXor;
  with Connection do  begin
    Canvas.MoveTo(Points[1].X, Points[1].Y);
    Canvas.LineTo(Points[2].X, Points[2].Y);
  end;
*)
end;

procedure TXCanvasForm.FormPaint(Sender: TObject);

  procedure PaintGrid;
  var
    X,Y: integer;
  begin
    X := 0;
    while X < Width do  begin
      Y := 0;
      while Y < Height do  begin
        {Canvas.Pixels[X,Y] := clBlack;}
        SetPixelV(Canvas.Handle, X,Y, clBlack);
        Inc(Y, Grid.Y);
      end;
      Inc(X, Grid.X);
    end;
  end;

begin
{ paint grid }
  PaintGrid;
end;

procedure TXCanvasForm.FormCreate(Sender: TObject);
begin
  Types := TStringList.Create;
  Connections := TStringList.Create;
  Grid := Point(10,10);
end;

procedure TXCanvasForm.FormMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);

  procedure DrawLine;
  begin
    Canvas.MoveTo(LastPoints[1].X, LastPoints[1].Y);
    Canvas.LineTo(LastPoints[2].X, LastPoints[2].Y);
  end;

  procedure MoveSchema(DeltaX, DeltaY: integer);
  var
    i: integer;
  begin
    for i := 0 to Types.Count-1 do  begin
      with TXplainType(Types.Objects[i]) do  begin
        Left := ((OrigLeft + DeltaX) div Grid.X) * Grid.X;
        Top := ((OrigTop + DeltaY) div Grid.Y) * Grid.Y;
      end;
    end;
  end;

begin
  case Mode of
    mConnectTypes:
      begin
        Canvas.Pen.Mode := pmNotXor;
        if HaveLastPoints then  begin
          DrawLine;
        end;
        LastPoints[1].X := SelectedType.Left + ConnectOrigin.X;
        LastPoints[1].Y := SelectedType.Top + ConnectOrigin.Y;
        LastPoints[2] := Point(X,Y);
        DrawLine;
        HaveLastPoints := True;
      end;
    mMoveSchema:
      begin
        if ClickOrigin.X = -1
          then  begin
            ClickOrigin.X := X;
            ClickOrigin.Y := Y;
            MovePoint.X := X;
            MovePoint.Y := Y;
          end
          else  begin
            if (((X mod Grid.X) * Grid.X) <> MovePoint.X) or
               (((Y mod Grid.Y) * Grid.Y) <> MovePoint.Y) then  begin
              MoveSchema(X-ClickOrigin.X, Y-ClickOrigin.Y);
              MovePoint.X := X;
              MovePoint.Y := Y;
            end;
          end;
      end;
  end; { of case }
end;

procedure TXCanvasForm.ClearLine;
begin
  if HaveLastPoints then  begin
    Canvas.Pen.Mode := pmNotXor;
    Canvas.MoveTo(LastPoints[1].X, LastPoints[1].Y);
    Canvas.LineTo(LastPoints[2].X, LastPoints[2].Y);
    HaveLastPoints := False;
  end;
end;

procedure TXCanvasForm.FormMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  i: integer;
  RegionPoints: array[1..4] of TPoint;
  HRgn: THandle;
begin
  case Mode of
    mAddTypes:
      begin
        for i := 0 to Connections.Count - 1 do  begin
          with TXPlainConnection(Connections.Objects[i]) do
            Selected := False;
        end;
        for i := 0 to Connections.Count - 1 do  begin
          with TXPlainConnection(Connections.Objects[i]) do  begin
            Selected := False;
            RegionPoints[1].X := X1-5;
            RegionPoints[1].Y := Y1;
            RegionPoints[2].X := X1+5;
            RegionPoints[2].Y := Y1;
            RegionPoints[3].X := X2+5;
            RegionPoints[3].Y := Y2;
            RegionPoints[4].X := X2-5;
            RegionPoints[4].Y := Y2;
            HRgn := CreatePolygonRgn(RegionPoints, 4, ALTERNATE);
            if HRgn <> 0 then
              try
                if PtInRegion(HRgn, X, Y) then  begin
                  Selected := True;
                  SelectedConnection := TXPlainConnection(Connections.Objects[i]);
                  break;
                end;
              finally
                DeleteObject(HRgn);
              end;
          end; { of with }
        end; { of for i }
      end;
    mMoveSchema:
      begin
        Mode := mAddTypes;
        Refresh;
      end;
  end; { of case }
  Refresh;
end;

procedure TXCanvasForm.Save1Click(Sender: TObject);
begin
  if FileName = ''
    then
      SaveAs1Click(Sender)
    else
      SaveTypeAsTexFile;
end;

procedure TXCanvasForm.SaveTypeAsTexFile;
var
  LowestX,
  LowestY,
  HighestX,
  HighestY: integer;

  procedure DetermineXAndYRange;
  var
    i: integer;
    X,Y: integer;
  begin
    LowestX := MaxInt;
    HighestX := 0;
    LowestY := MaxInt;
    HighestY := 0;
    for i := 0 to Types.Count - 1 do  begin
      with TXplainType(Types.Objects[i]) do  begin
        if Left < LowestX then
          LowestX := Left;
        X := Left+Width;
        if X > HighestX then
          HighestX := X;
        if Top < LowestY then
          LowestY := Top;
        Y := Top+Height;
        if Y > HighestY then
          HighestY := Y;
      end;
    end;
  end;

  procedure WriteLaTeXFigure;
  var
    ftex: System.text;

    procedure WriteTypes;
    var
     i: integer;
    begin
      for i := 0 to Types.Count - 1 do  begin
        with TXplainType(Types.Objects[i]) do  begin
          writeln(ftex, '\put(', Left, '.00,', HighestY-Top-Height, '.00)',
                        '{\framebox(', Width, '.00,', Height, '.00)[cc]',
                        '{', TypeFormat, Caption, '}}}');
        end;
      end;
    end;

    procedure WriteConnections;

      function FindGreatestCommonDivisor(A,B: integer): integer;
      begin
        A := Abs(A);
        B := Abs(B);
        if A > B
          then
            Result := B div 2
          else
            Result := A div 2;
        while (Result > 1) and
              ((A mod Result <> 0) or (B mod Result <> 0)) do
          Dec(Result)
      end;

    var
     i: integer;
     VerticalLine,
     HorizontalLine: Boolean;
     PointNumber: cardinal;
    begin
      PointNumber := 1;
      for i := 0 to Connections.Count - 1 do  begin
        with TXPlainConnection(Connections.Objects[i]) do  begin
          write(ftex, '\put(', X1, '.00,', HighestY-Y1, '.00)');
          VerticalLine := X1 - X2 = 0;
          if VerticalLine
            then  begin
              if Y2 > Y1
                then
                  writeln(ftex, '{\line(0,-1){', Abs(Y2 - Y1), '}}')
                else
                  writeln(ftex, '{\line(0,1){', Abs(Y2 - Y1), '}}');
            end
            else  begin
              HorizontalLine := Y1 - Y2 = 0;
              if HorizontalLine
                then  begin
                  if X2 >= X1
                    then
                      writeln(ftex, '{\line(1,0){', Abs(X2-X1), '}}')
                    else
                      writeln(ftex, '{\line(-1,0){', Abs(X2-X1), '}}');
                end
                else  begin
                  writeln(ftex, '{\emline{0}{0}',
                                        '{', PointNumber, '}',
                                        '{', X2-X1, '}{', Y1-Y2, '}',
                                        '{', PointNumber+1, '}}');
                  Inc(PointNumber, 2);
                end;
            end;
        end;
      end;
    end;

  begin
    AssignFile(ftex, FileName);
    Rewrite(ftex);
    writeln(ftex, '\unitlength 0.30mm');
    writeln(ftex, '\special{em:linewidth 0.4pt}');
    writeln(ftex, '\linethickness{0.4pt}');
    writeln(ftex, '\begin{picture}(', HighestX, '.00,', HighestY-LowestY, '.00)',
                                 '(', LowestX, '.00,0.00)');
    WriteTypes;
    WriteConnections;
    writeln(ftex, '\end{picture}');
    CloseFile(ftex);
  end;

  procedure WriteEPS;
  var
    feps: System.text;

    procedure WriteTypes;
    var
     i: integer;
    begin
      for i := 0 to Types.Count - 1 do  begin
        with TXplainType(Types.Objects[i]) do  begin
          writeln(feps, Left-LowestX, ' ', ((HighestY-LowestY)-Top)+1, ' moveto');
          writeln(feps, (Left-LowestX)+Width, ' ', ((HighestY-LowestY)-Top)+1, ' lineto');
          writeln(feps, (Left-LowestX)+Width, ' ', ((HighestY-LowestY)-Top)+1-Height, ' lineto');
          writeln(feps, (Left-LowestX), ' ', ((HighestY-LowestY)-Top)+1-Height, ' lineto');
          writeln(feps, 'closepath');
          writeln(feps, 'stroke');
          writeln(feps, '(', Caption, ')');
          writeln(feps, 'dup stringwidth');
          writeln(feps, Height, ' exch sub 2 div ', ((HighestY-LowestY)-Top)+1-Height, ' add');
          writeln(feps, 'exch');
          writeln(feps, Width, ' exch sub 2 div ', Left-LowestX, ' add');
          writeln(feps, 'exch moveto show');
        end;
      end;
    end;

    procedure WriteConnections;
    var
      i: integer;
    begin
      for i := 0 to Connections.Count - 1 do  begin
        with TXPlainConnection(Connections.Objects[i]) do  begin
          writeln(feps, X1-LowestX, ' ', ((HighestY-LowestY)-Y1)+1, ' moveto');
          writeln(feps, X2-LowestX, ' ', ((HighestY-LowestY)-Y2)+1, ' lineto stroke');
        end; { of with }
      end;
    end;

  begin
    AssignFile(feps, ChangeFileExt(FileName, '.eps'));
    Rewrite(feps);
    writeln(feps, '%!PS-Adobe-3.0 EPSF-3.0');
    writeln(feps, '%%Creator: (dplain)');
    writeln(feps, '%%BoundingBox: 0 0 ', HighestX-LowestX, ' ', (HighestY-LowestY)+1);
    writeln(feps, '%%Pages: 0');
    writeln(feps, '%%EndComments');
    writeln(feps, '0.75 0.75 scale');
    writeln(feps, '0.4 setlinewidth');
    writeln(feps, '/Helvetica findfont');
    writeln(feps, '11 scalefont');
    writeln(feps, 'setfont');
    WriteTypes;
    WriteConnections;
    writeln(feps, '%%PageTrailer');
    writeln(feps, '%%Trailer');
    writeln(feps, '%%BoundingBox: ', LowestX, ' ', LowestY, ' ', HighestX, ' ', HighestY-LowestY);
    writeln(feps, '%%EOF');
    CloseFile(feps);
  end;

  procedure WriteMetaPost;
  var
    fmp: System.text;

    procedure WriteTypes;
    var
     i: integer;

     function BreakWordsIntohbox(const s: string): string;
     var
       first,
       i: integer;
     begin
       if Pos(' ', s) <> 0
         then  begin
           first := 1;
           Result := '';
           for i := 1 to length(s) do  begin
             if s[i] = ' ' then  begin
               Result := Result + '\hbox{' + Copy(s, first, i-first) + '}';
               first := i+1;
             end;
           end;
           Result := Result + '\hbox{' + Copy(s, first, length(s)-first+1) + '}';
         end
         else  begin
           Result := '\hbox{' + s + '}';
         end;
     end;

    begin
      for i := 0 to Types.Count - 1 do  begin
        with TXplainType(Types.Objects[i]) do  begin
          writeln(fmp, 'q := p shifted (', Left-LowestX, ',', ((HighestY-LowestY)-Top-DefaultHeight),'); fill q withcolor white; draw q;');
          writeln(fmp, 'label(btex \vbox{', BreakWordsIntohbox(Caption), '} etex, center q);');
        end;
      end;
    end;

    procedure WriteConnections;
    var
      i: integer;
      FromX: integer;
      p1,q1,p2,q2,p3,q3: integer;
    begin
      for i := 0 to Connections.Count - 1 do  begin
        with TXPlainConnection(Connections.Objects[i]) do  begin
          if IsSubType
            then
              FromX := (X1-LowestX)
            else
              FromX := (X1-LowestX);
          if IsSelfReference
            then  begin
              p1 := X1;
              q1 := Y1;
              p3 := X2;
              q3 := Y2;
              if X1ConnectionPoint <= 2
                then  p2 := Type1.Left - (p1-Type1.Left)
                else  p2 := (Type1.Left + Type1.Width) + ((Type1.Left + Type1.Width) - p1);
              if q3 > q1
                then  q2 := q1 + ((q3 - q1) div 2)
                else  q2 := q3 + ((q1 - q3) div 2);
              writeln(fmp, 'draw (', p1-LowestX, ',', ((HighestY-LowestY)-q1), ')..',
                           '(', p2-LowestX, ',', ((HighestY-LowestY)-q2), ')..',
                           '(', p3-LowestX, ',', ((HighestY-LowestY)-q3), ');');
            end
            else  begin
              writeln(fmp, 'draw (', FromX, ',', ((HighestY-LowestY)-Y1), ')--',
                           '(', X2-LowestX, ',', ((HighestY-LowestY)-Y2), ');');
            end;
        end; { of with }
      end;
    end;

  begin
    AssignFile(fmp, ChangeFileExt(FileName, '.mp'));
    Rewrite(fmp);
    writeln(fmp, 'input mp-tool;');
    writeln(fmp, 'verbatimtex \input dplain etex');
    writeln(fmp);
    writeln(fmp, 'beginfig(1);');
    writeln(fmp, 'path p, q ; p := unitsquare xscaled 90 yscaled 40;');
    //writeln(fmp, 'defaultfont:="cmss10";');
    WriteTypes;
    WriteConnections;
    writeln(fmp, 'endfig;');
    writeln(fmp, 'end;');
    CloseFile(fmp);
  end;

begin
  DetermineXAndYRange;
  if FileExists(FileName) then  begin
    DeleteFile(FileName + '.bak');
    RenameFile(FileName, FileName + '.bak');
  end;
  WriteLaTeXFigure;
  {WriteEPS;}
  WriteMetaPost;
  Changed := False;
end;

procedure TXCanvasForm.OpenTexFile;
var
  ftex: System.text;
  Line: string;
  LowestX,
  LowestY,
  HighestX,
  HighestY: integer;

  function FTCopy(const From: string; First, Last: integer): string;
  begin
    Result := Copy(From, First, Last-First+1);
  end;

  function ParsePoint(Start: cardinal): TPoint;
  { parses a (X,Y) structure which starts at Start }
  var
    p,q: integer;
  begin
    p := Start;
    q := p+1;
    while not (Line[q] in ['.',',']) do
      Inc(q);
    Result.X := StrToInt(FTCopy(Line, p+1, q-1));
    if Line[q] = '.'
      then  begin
        p := q + 1;
        while Line[p] <> ',' do
          Inc(p);
      end
      else
        p := q;
    q := p+1;
    while not (Line[q] in [')','.']) do
      Inc(q);
    Result.Y := StrToInt(FTCopy(Line, p+1, q-1));
  end;

  function ParsePut: TPoint;
  begin
    ParsePut := ParsePoint(Pos('(', Line));
  end;

  procedure ParseType;
  var
    A,B: TPoint;
    p: integer;
    AType: TXplainType;
    TypeName: string;
  begin
    A := ParsePut;
    p := Pos('\framebox', Line);
    B := ParsePoint(p+length('\framebox'));
    while Line[p] <> '{' do
      Inc(p);
    TypeName := FTCopy(Line, p+1, length(Line)-2);
    if Pos(TypeFormat, TypeName) = 1 then
      TypeName := FTCopy(TypeName, length(TypeFormat)+1, length(TypeName)-1);
    if Pos('\typeformat{', TypeName) = 1 then
      TypeName := FTCopy(TypeName, length('\typeformat{')+1, length(TypeName)-1);
    AType := TXplainType.Create(Self);
    AType.Left := A.X;
    AType.Top := HighestY - B.Y - A.Y;
    AType.Width := B.X;
    AType.Height := B.Y;
    AType.Caption := TypeName;
    AType.Parent := Self;
    AType.OnMouseDown := Panel1MouseDown;
    AType.OnMouseMove := Panel1MouseMove;
    AType.OnMouseUp := Panel1MouseUp;
    AType.PopUpMenu := TypePopupMenu;
    Types.AddObject(AType.Caption, AType);
  end;

  procedure ConnectTypes(A,B: TPoint);
  var
    Type1,
    Type2: TXplainType;

    procedure FindTypesWhichLineTouches;

      function FindTypeWhichEndPointTouches(Point: TPoint): TXplainType;

        function TouchesType(Point: TPoint; AType: TXplainType): Boolean;
        begin
          Result := False;
          if (Point.Y = AType.Top) or (Point.Y = AType.Top+AType.Height) then  begin
            if (Point.X >= AType.Left) and (Point.X <= AType.Left+AType.Width) then  begin
              Result := True;
              Exit;
            end;
          end;
        end;

      var
        i: integer;
      begin
        Result := nil;
        for i := 0 to Types.Count-1 do  begin
          if TouchesType(Point, TXplainType(Types.Objects[i])) then  begin
            Result := TXplainType(Types.Objects[i]);
            break;
          end;
        end;
      end;

    begin
      Type1 := FindTypeWhichEndPointTouches(A);
      Type2 := FindTypeWhichEndPointTouches(B);
      if (Type1 = nil) or (Type2 = nil) then
        raise EInvalidPictureFile.Create('Error in picture file. Unconnected line found.');
    end;

  begin
    A.Y := HighestY - A.Y;
    B.Y := HighestY - B.Y;
    try
      FindTypesWhichLineTouches;
      A := Type1.ScreenToClient(ClientToScreen(A));
      B := Type2.ScreenToClient(ClientToScreen(B));
      Type1.ConnectTo(Type2, A.X, A.Y, B.X, B.Y);
    except
    end;
  end;

  procedure ParseLineConnection;
  var
    A,B: TPoint;
    HorizontalLine: Boolean;
    p: integer;
    Len: integer;
  begin
    A := ParsePut;
    p := Pos('\line', Line);
    B := ParsePoint(p+length('\line'));
    HorizontalLine := B.Y = 0;
    p := length(Line)-2;
    while Line[p] <> '{' do
      Dec(p);
    Len := StrToInt(FTCopy(Line, p+1, length(Line)-2));
    if HorizontalLine
      then  begin
      { horizontal line }
        B.X := A.X + B.X * Len;
        B.Y := A.Y;
      end
      else  begin
      { vertical line }
        B.X := A.X;
        B.Y := A.Y + B.Y * Len;
      end;
    ConnectTypes(A,B);
  end;

  procedure ParseEmLineConnection;
  var
    A,B: TPoint;
    p,q: cardinal;
  begin
    A := ParsePut;
    p := Pos('\emline{', Line) + length('\emline{')+1;
    while Line[p] <> '{' do
      Inc(p);
    p := p+1;
    while Line[p] <> '{' do
      Inc(p);
    p := p+1;
    while Line[p] <> '{' do
      Inc(p);
    q := p+1;
    while Line[q] <> '}' do
      Inc(q);
    B.X := StrToInt(FTCopy(Line, p+1, q-1));
    p := q+1;
    q := p+1;
    while Line[q] <> '}' do
      Inc(q);
    B.Y := StrToInt(FTCopy(Line, p+1, q-1));
    B.X := A.X + B.X;
    B.Y := A.Y + B.Y;
    ConnectTypes(A,B);
  end;


  procedure ParsePictureDimensions;
  var
    A,B: TPoint;
    p: integer;
  begin
    p := Pos('(', Line);
    A := ParsePoint(p);
    HighestX := A.X;
    HighestY := A.Y;
    Inc(p);
    while (p <= length(Line)) and (Line[p] <> '(') do
      Inc(p);
    if p < length(Line)
      then  begin
        B := ParsePoint(p);
        LowestX := B.X;
        LowestY := B.Y;
      end
      else  begin
        LowestX := 0;
        LowestY := 0;
      end;
  end;

begin
  AssignFile(ftex, FileName);
  Reset(ftex);
  while not Eof(ftex) do  begin
    readln(ftex, Line);
    if Pos('\framebox(', Line) <> 0
      then  ParseType
      else  if Pos('\line(', Line) <> 0
      then  ParseLineConnection
      else  if Pos('\emline{', Line) <> 0
      then  ParseEmLineConnection
      else if Pos('\begin{picture}', Line) <> 0
      then  ParsePictureDimensions;
  end;
  CloseFile(ftex);
end;

procedure TXCanvasForm.SaveAs1Click(Sender: TObject);
var
  Dialog: TSaveDialog;
begin
  Dialog := TSaveDialog.Create(Self);
  try
    Dialog.DefaultExt := 'tex';
    Dialog.Filter := 'TeX files (*.tex)|*.tex|Any file (*.*)|*.*';
    if FileName = ''
      then
        Dialog.FileName := '*.tex'
      else
        Dialog.FileName := ExtractFileName(FileName);
    if Dialog.Execute then
      FileName := Dialog.FileName;
  finally
    Dialog.Free;
  end;
  if FileName <> '' then
    SaveTypeAsTexFile;
end;

procedure TXCanvasForm.Open1Click(Sender: TObject);
var
  Dialog: TOpenDialog;
begin
  Dialog := TOpenDialog.Create(Self);
  try
    Dialog.DefaultExt := 'tex';
    Dialog.Filter := 'TeX files (*.tex)|*.tex|Any file (*.*)|*.*';
    Dialog.FileName := '*.tex';
    if not Dialog.Execute then
      Exit;
    New1Click(Self);
    FileName := Dialog.FileName;
    OpenTexFile;
  finally
    Dialog.Free;
  end;
end;

procedure TXCanvasForm.New1Click(Sender: TObject);
var
  i: integer;
begin
  if Changed then
  begin
    case UserWants('Schema changed. Save it?') of
      mrCancel: Exit;
      mrYes: Save1Click(Sender);
      mrNo: { don't save }
    end;
  end;
  for i := 0 to Types.Count-1 do
    Types.Objects[i].Free;
  Types.Clear;
  for i := 0 to Connections.Count-1 do
    Connections.Objects[i].Free;
  Connections.Clear;
  SelectedType := nil;
  SelectedConnection := nil;
  ClickedType := nil;
  DraggedType := nil;
  Mode := mAddTypes;
  Changed := False;
  FileName := '';
  Refresh;
end;

procedure TXCanvasForm.Move1Click(Sender: TObject);
var
  i: integer;
begin
  Mode := mMoveSchema;
  ClickOrigin.X := -1;
  ClickOrigin.Y := -1;
  for i := 0 to Types.Count-1 do  begin
    with TXplainType(Types.Objects[i]) do  begin
      OrigLeft := Left;
      OrigTop := Top;
    end;
  end;
end;


procedure TXCanvasForm.SetSelectedType(AType: TXplainType);
begin
  if SelectedType <> nil then  begin
    SelectedType.Font.Color := clBlack;
  end;
  SelectedType := AType;
  if SelectedType <> nil then
    SelectedType.Font.Color := clGreen;
end;


procedure TXCanvasForm.RemoveConnection(AConnection: TXPlainConnection);
begin
  Connections.Delete(Connections.IndexOfObject(AConnection));
  AConnection.RemoveLinks;
  if AConnection = SelectedConnection then
    SelectedConnection := nil;
  AConnection.Free;
  Changed := True;
end;


procedure TXCanvasForm.RemoveType(AType: TXplainType);
var
  i: integer;
begin
{ remove all connections to this type }
  i := 0;
  while  i < Connections.Count do  begin
    with TXPlainConnection(Connections.Objects[i]) do  begin
      if (Type1 = AType) or (Type2 = AType)
        then  begin
          RemoveConnection(TXPlainConnection(Connections.Objects[i]));
        end
        else
          Inc(i);
    end;
  end;

{ remove the type itself }
  Types.Delete(Types.IndexOfObject(AType));
  if AType = SelectedType then
    SetSelectedType(nil);
  AType.Free;
  Changed := True;
end;


procedure TXCanvasForm.SetFileName(AFileName: string);
begin
  FFileName := AFileName;
  if FileName = ''
    then
      Caption := 'DPlain - <new schema>'
    else
      Caption := 'DPlain - ' + FFileName;
end;


procedure TXCanvasForm.FormActivate(Sender: TObject);
var
  r: record
       case integer of
         0: (c: array[1..8] of byte);
         1: (d: double);
     end;
begin
  r.c[8] := $3F;
  r.c[7] := $EF;
  r.c[6] := $40;
  r.c[5] := $00;
  r.c[4] := $00;
  r.c[3] := $00;
  r.c[2] := $00;
  r.c[1] := $00;
end;

procedure TXCanvasForm.Movetoback1Click(Sender: TObject);
begin
  if Assigned(SelectedType) then
    MoveBehind(SelectedType);
end;


procedure TXCanvasForm.MoveBehind(AType: TXplainType);
begin
{require}
  Assert(Assigned(AType), 'No type assigned');

{ remove the type itself }
  Types.Delete(Types.IndexOfObject(AType));
  Types.InsertObject(0, AType.Caption, AType);
  AType.SetZOrder(False);
  Changed := True;
end;


procedure TXCanvasForm.FormCloseQuery(Sender: TObject;
  var CanClose: Boolean);
begin
  if Changed then  begin
    case MessageDlg('Schema changed. Save changes?', mtConfirmation, mbYesNoCancel, 0) of
      mrYes: Save1Click(Sender);
      mrCancel: CanClose := False;
    else {}
    end;
  end;
end;

procedure TXCanvasForm.Delete1Click(Sender: TObject);
begin
  if (SelectedType <> nil) and
     (UserWants('Delete this type?') = mrYes) then
  begin
    RemoveType(SelectedType);
    Refresh;
  end;
end;


end.
